#!/usr/bin/env python
"""
Created 28 Jul 2020
This script creates a png image of the contents of a SIR file using
default vmin/vmax values in file header

@author: DGL at BYU
"""

######
# Imports
######
import os
import numpy as np
from sir_io import *   # sirpi io module
import png as png      # PyPng module
from sir2png import *
from sir_geom import ease2_map_info

def main(argv): # pass in a filename as argument
    vmin=0
    vmax=0
    if len(argv)>1:
        sir_fname = argv[1]
        if len(argv)>2:
            tiff_fname = argv[2]
            if len(argv)>4:
                vmin=float(argv[3])
                vmax=float(argv[4])
        else:
            tiff_fname = sir_fname + '.tiff'
    else: # no argument?, use a test image
        sir_fname = 'greeni.sir'
        #sir_fname = 'queh-a-E2N04-10-10.sir'
        tiff_fname = sir_fname + '.tiff'

    # create a simple .png image
    png_fname = '/tmp/' + sir_fname + '_tmp.png'
    print('Create temporary .png file '+png_fname)
    sir2png(sir_fname,png_fname,vmin=vmin,vmax=vmax)

    # read sir file head
    #print('Create gdal_translate string')
    sir = loadsir(sir_fname,headonly=1) # read only header array
    head=sir[1]

    # extract projection info from head
    iopt=sirheadvalue('iopt',head)
    nsx=sirheadvalue('nsx',head)
    nsy=sirheadvalue('nsy',head)
    a0=sirheadvalue('a0',head)
    b0=sirheadvalue('b0',head)
    xdeg=sirheadvalue('xdeg',head)
    ydeg=sirheadvalue('ydeg',head)
    ascale=sirheadvalue('ascale',head)
    bscale=sirheadvalue('bscale',head)

    # create argument list to gdal_translate
    gd=None # default, do nothing
    if iopt == -1: 
        print("Image-only file -- no map projection information");

    elif iopt == 0:
        print("Rectangular Lat/Lon projection");
        gd='-of GTiff -a_srs "+proj=eqc +lat_ts=0 +lat_0=0 +lon_0=0 +x_0=0 +y_0-0 +ellips=WGS84 +units=m" -a_ullr {:0.1f} {:0.1f} {:0.1f} {:0.1f} '.format(a0,b0,a0+xdeg,b0+ydeg)
        
    elif iopt == 1 or iopt == 2:
        print("Lambert Equal Area Azimuth projection");
        Erad=6378.0
        if iopt==2:
            F=298.26
            era=(1.-1./F)
            Erad=Erad*era/np.sqrt(np.square(era*np.cos(np.deg2rad(ydeg)))+np.square(np.sin(np.deg2rad(ydeg))))
        x0=a0
        x1=a0+(nsx-1)/ascale
        y0=b0
        y1=b0+(nsy-1)/bscale
        gd='-of GTiff -a_srs "+a={:0.3f} +rf=50000000 +proj=laea +lat_0={:0.3f} +lon_0={}" -a_ullr {:0.1f} {:0.1f} {:0.1f} {:0.1f} '.format(Erad*1000,ydeg,xdeg,x0*1000,y1*1000,x1*1000,y0*1000)

    elif iopt==5:
        Erad=6378.273
        x0=a0
        x1=a0+(nsx-1)*ascale
        y0=b0
        y1=b0+(nsy-1)*bscale
        print("Polar Sterographic projection")
        lat0 = 90.0
        if ydeg < 0.0:
            lat0 = -90.0
        gd='-of GTiff -a_srs "+datum=wgs84 +proj=stere +lat_0={:0.3f} +lat_ts={:0.3f} +lon_0={:0.3f} +k=1 +x_0=0 +y_0=0 +a=6378273.000 +b=6356889.449 +units=m +no_defs" -a_ullr {:0.3f} {:0.3f} {:0.3f} {:0.3f} '.format(lat0,ydeg,xdeg,x0*1000,y1*1000,x1*1000,y0*1000)

    elif iopt==8 or iopt==9:
        print("EASE2 grid polar projection (lambert_azimuthal_equal_area): %d" % iopt)
        lat0=90.0
        if iopt==9:
            lat0 = -90.0

        # get base EASE2 map projection parameters
        ind = np.round(bscale)
        isc = np.round(ascale)
        (map_equatorial_radius_m, map_eccentricity, \
         e2, map_reference_latitude, map_reference_longitude, \
         map_second_reference_latitude, sin_phi1, cos_phi1, kz, \
         map_scale, bcols, brows, r0, s0, epsilon) = ease2_map_info(iopt, isc, ind)
        x0=-map_scale*s0
        y0=-map_scale*r0
        x1=map_scale*s0
        y1=map_scale*r0     
        gd='-of GTiff -a_srs "+proj=laea +lat_0={:0.3f} +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m" -a_ullr {:0.3f} {:0.3f} {:0.3f} {:0.3f} '.format(lat0,x0,y1,x1,y0)

    elif iopt==10:
        print("EASE2 grid global projection (lambert_cylindrical_equal_area): %d" % iopt)
        # get base EASE2 map projection parameters
        ind = np.round(bscale)
        isc = np.round(ascale)
        (map_equatorial_radius_m, map_eccentricity, \
         e2, map_reference_latitude, map_reference_longitude, \
         map_second_reference_latitude, sin_phi1, cos_phi1, kz, \
         map_scale, bcols, brows, r0, s0, epsilon) = ease2_map_info(iopt, isc, ind)
        x0=-map_scale*s0
        y0=-map_scale*r0
        x1=map_scale*s0
        y1=map_scale*r0     
        gd='-of GTiff -a_srs "+proj=cea +lat_0=0 +lon_0=0 +lat_ts=30 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m " -a_ullr {:0.3f} {:0.3f} {:0.3f} {:0.3f} '.format(x0,y1,x1,y0)

    elif iopt==11 or iopt==12:
        print("EASE Polar Azimuthal projection");
        printf("*gdal_translate does not support this projection")

    elif iopt==13:  
        print("EASE Cylindrical projection")
        printf("*gdal_translate does not support this projection")
    else: 
        print("Unknown (and unsupported) projection:")
        print(" Image size: (%d X %d)" % (nsx,nsy))

    # create geotiff file using gdal_translate
    print('Using gdal_translate to create .geotiff file '+tiff_fname)
    #print(gd)
    gdal_translate_path='gdal_translate' # full path to gdal_translate
    os.system(gdal_translate_path + " " + gd + png_fname + " " + tiff_fname)

    # delete temporary .png file
    os.system("rm " + png_fname)

if __name__ == '__main__':
    main(sys.argv)
